///////////////////////////////////////////////////
// Calculating the total aquisition number (weighted version, Poisson distribution)
// Author: Pawel Pralat
// Last modification: November 25, 2015
///////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

// n : the max length of the path
const int n = 21;

// the max weight generated, path of length i deals with weithgs at most k[i] 
const int max_k = 10;
int k[n+1] = {0, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 4, 4, 4, 3, 3, 3};

// table containing weights
int t[n];

// conditional Poisson distribution: p[i] is the probability Po(1) = i (conditioning on Po(1) \ge 1)
double p[max_k+1]; 

// q[i] is the probability that the "island" has length i
double q[n+1]; 

double expectation;
double remaining_prob;

///////////////////////////////////////////////////
// Testing if interval [a, b] has weithgs that can be moved to a single vertex
///////////////////////////////////////////////////
int test(int a, int b)
{
  int l = a;
  int tmp = t[a];
  while ( (l < b) && (tmp <= t[l+1]) )
    {
      tmp += t[l+1];
      l++;
    }

  int r = b;
  tmp = t[b];
  while ( (r > a) && (tmp <= t[r-1]) )
    {
      tmp += t[r-1];
      r--;
    }

  if ( r-l <= 1 )
    return 1;
  else 
    return 0;
}

///////////////////////////////////////////////////
// Checking weighted path of length m
///////////////////////////////////////////////////
int acq(int m)
{
  int i = 0;
  int nr = 0;

  while (i < m)
    {
      int j = 0;
      while ( (test(i,i+j)) && (i+j < m) )
        j++;

      i += j;
      nr++;
    }

  return nr;
}

///////////////////////////////////////////////////
// Generating weighted paths of length m with max weight l; vertex i is the next to fix
///////////////////////////////////////////////////
void generating(int i, int l, int m)
{
  if (i == m) // path is ready!
    {
      double prob = 1;
      for (int j = 0 ; j < m ; j++)
	{
	  prob *= p[t[j]];
	  // printf("%d ", t[j]);
	}
      // printf("-- %d %f \n", acq(m), prob);
      remaining_prob -= prob;
      expectation += prob*acq(m);
      return;
    }

  for (int j = 1 ; j <= l ; j++)
    {
      t[i] = j;
      generating(i+1, l, m);
    }
}

///////////////////////////////////////////////////
// Main program
///////////////////////////////////////////////////
int main(int argc, char * argv[])
{
  // Generating Conditional Poisson distribution
  p[1] = 1.0/(exp(1.0)-1.0);
  for (int i = 2 ; i <= max_k ; i++)
    p[i] = p[i-1]/(1.0*i);

  // Generating Geometric distributions for q
  q[1] = (1.0 - 1.0/exp(1.0))/exp(1.0);
  for (int i = 2 ; i <= n ; i++)
    q[i] = q[i-1]*(1.0-1.0/exp(1.0));

  double lower_bound = q[1]+q[2]; 
  double upper_bound = q[1]+q[2];

  for (int i = 3; i <= n ; i++)
    {
      remaining_prob = 1.0;
      expectation = 0.0;
      generating(0,k[i],i);
      lower_bound += q[i] * expectation;
      upper_bound += q[i] * (expectation + ceil(i/2.0)*remaining_prob);

      time_t rawtime;
      struct tm * timeinfo;

      time ( &rawtime );
      timeinfo = localtime ( &rawtime );

      printf("n=%2d k=%2d lb=%17.15f ub=%17.15f ub-lb=%e prob=%e diff=%e time=%s", i, k[i], expectation, expectation + ceil(i/2.0)*remaining_prob, ceil(i/2.0)*remaining_prob, q[i], ceil(i/2.0)*remaining_prob*q[i]/(exp(1.0)), asctime (timeinfo));
    }

  double r = q[n];
  for (int i = n+1 ; i <= 500 ; i++)
    {
      r *= (1.0-1.0/exp(1.0));
      upper_bound += r*ceil(i/2.0);
      lower_bound += r*expectation;
    }
  lower_bound /= exp(1.0);
  upper_bound /= exp(1.0);

  printf("LOWER BOUND = %20.15f \n", lower_bound);
  printf("UPPER BOUND = %20.15f \n", upper_bound);

  return 0;
}
